use serde::Serialize;
use sha2::{Digest, Sha256};
use std::fmt::Display;

#[derive(thiserror::Error, Debug)]
pub enum Error {
    #[error("Invalid length for {0}: {1} (expected: {2})")]
    InvalidLength(String, usize, usize),
}

macro_rules! Id {
    ($id:ident, $suffix:expr) => {
        #[derive(Clone, Debug, Eq, Hash, PartialEq, Serialize)]
        pub struct $id {
            bytes: [u8; 32],
        }

        impl $id {
            pub fn new(bytes: [u8; 32]) -> Self {
                Self { bytes }
            }
        }

        impl std::str::FromStr for $id {
            type Err = Error;
            fn from_str(value: &str) -> Result<Self, Self::Err> {
                Ok(Self::try_from(value).unwrap_or_else(|_| {
                    let mut hasher = Sha256::new();
                    hasher.update(&value);
                    Self {
                        bytes: hasher.finalize().into(),
                    }
                }))
            }
        }

        impl TryFrom<&str> for $id {
            type Error = Error;
            fn try_from(value: &str) -> Result<Self, Self::Error> {
                if $suffix.len() + 64 != value.len() {
                    return Err(Error::InvalidLength(
                        stringify!($id).to_owned(),
                        value.len(),
                        $suffix.len() + 64,
                    ));
                }

                if !value.starts_with($suffix) {
                    return Err(Error::InvalidLength(
                        stringify!($id).to_owned(),
                        value.len(),
                        $suffix.len() + 64,
                    ));
                }

                let bytes = hex::decode(&value[$suffix.len()..]).map_err(|_| {
                    Error::InvalidLength(stringify!($id).to_owned(), value.len(), 32)
                })?;
                bytes.try_into()
            }
        }

        impl TryFrom<Vec<u8>> for $id {
            type Error = Error;

            fn try_from(value: Vec<u8>) -> Result<Self, Self::Error> {
                if value.len() != 32 {
                    return Err(Error::InvalidLength(
                        stringify!($id).to_owned(),
                        value.len(),
                        32,
                    ));
                }
                let mut bytes = [0u8; 32];
                bytes.copy_from_slice(&value);
                Ok(Self { bytes })
            }
        }

        impl Display for $id {
            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                write!(f, "{}{}", $suffix, hex::encode(self.bytes))
            }
        }

        impl AsRef<[u8]> for $id {
            fn as_ref(&self) -> &[u8] {
                &self.bytes
            }
        }

        impl AsRef<[u8; 32]> for $id {
            fn as_ref(&self) -> &[u8; 32] {
                &self.bytes
            }
        }
    };
}

Id!(AccountId, "account");
Id!(TransactionId, "tx");
